
/* ----------------------------------------------------------------------
   The SINUS Firewall -- a TCP/IP packet filter for Linux

   Written within the SINUS project at the University of Zurich,
   SWITCH, Telekurs Payserv AG, ETH Zurich.
   originally based on the sf Firewall Software (C) 1996 by Robert
   Muchsel and Roland Schmid.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

   SINUS Firewall resources:
   SINUS Homepage: http://www.ifi.unizh.ch/ikm/SINUS/
   Firewall Homepage: http://www.ifi.unizh.ch/ikm/SINUS/firewall.html
   Frequently asked questions: http://www.ifi.unizh.ch/ikm/SINUS/sf_faq.html
   Mailing list for comments, questions, bug reports: firewall@ifi.unizh.ch
   ----------------------------------------------------------------------  */



package sfclasses;

import java.util.*;
import java.awt.*;
import java.io.*;
import java.awt.event.*;

/**
 * This class implements address and service macros. An object of this
 * class holds all information about one macro. The static parts of
 * the class take care of the macro vectors and of static macros. 
 * Static macros are stored in the vector <i>staticMacros</i> and are
 * inserted into that vector in a static initializer. Dynamic macros
 * are stored in a vector in the ManageDomain class.
 * @version 1.0 10 Jan 1997
 * @author Roland E. Schmid
 */

public class Macro implements Persistent {

  /**
   * Initialize empty macro. This is used when loading data from a
   * persistent input stream.
   */
  public Macro() {
  }

  /**
   * Initialize macro.
   * @param number Macro number
   */
  public Macro(int number) {
    macroNumber = number;
  }

  /**
   * Open a dialog to edit the macro.
   * @param co Calling object, will be notified when the edit dialog is closed.
   * @param p Parent frame
   * @param md ManageDomain object holding the configuration data
   * @param newMacro The macro is being created and not changed. If the user
   * cancels the operation, the macro will be deleted from the vector.
   * @see Refreshable
   */
  public void edit(Refreshable co, Frame p, ManageDomain md, boolean newMacro) {
    MacroDialog d = new MacroDialog(co, p, this, md, newMacro);
    d.setVisible(true);
    if (macroNumber <= Rule.SF_FIRSTMACRO && macroNumber > firstDynamic)
      UserDialog.WarningBox("Static macro. Changes will not be saved!");
  }

  // persistence

  /**
   * Write object data to a persistent output stream
   * @param ps Stream
   * @see PersistentOutputStream
   */
  public void write(PersistentOutputStream ps) {
    ps.writeInt("Macro=", macroNumber);
    ps.writeString("Name=", macroName);
    ps.writeInt("Type=", macroType);
    ps.writeInt("Port=", port);
    ps.writeInt("Prend=", prend);
    ps.writePersistent("Addresses=", addresses);
  }

  /**
   * Read object data from a persistent input stream
   * @param ps Stream
   * @see PersistentInputStream
   */
  public void read(PersistentInputStream ps) throws java.io.IOException {
    macroNumber = ps.readInt("Macro=");
    macroName = ps.readString("Name=");
    macroType = ps.readInt("Type=");
    port = ps.readInt("Port=");
    prend = ps.readInt("Prend=");
    addresses = (AddrList)ps.readPersistent("Addresses=");
  }


  // data

  public int macroNumber;

  public String macroName = new String();

  public int macroType = MACRO_ADDRESSLIST;
  /**
   * macroType: use addresses 
   */
  public static final int MACRO_ADDRESSLIST = 1; 
  /**
   * macroType: inside 
   */
  public static final int MACRO_INSIDE      = 2;
  /**
   * macroType: outside
   */
  public static final int MACRO_OUTSIDE     = 3;
  /**
   * macroType: check for ports only
   */
  public static final int MACRO_PORTS_ONLY  = 4;
  /**
   * macroType: own addresses of the firewall 
   */
  public static final int MACRO_OWNADDR     = 5;

  public int macroSpecial = 0;
  /**
   * macroSpecial: insert addresses of all firewalls
   */
  public static final int MACRO_FIREWALLS   = 1; // firewall addresses inserted automatically
  /**
   * macroSpecial: insert addresses of all authorized configuration clients
   */
  public static final int MACRO_CONFCLIENTS = 2; // configuration client addresses inserted automatically
  /**
   * macroSpecial: macro is generated automatically from server template
   */
  public static final int MACRO_TEMPLATE    = 3; // macro generated automatically from server template

  public int port  = MACRO_FIRSTPORT;
  public int prend = MACRO_LASTPORT;
  public static final int MACRO_FIRSTPORT = 0;
  public static final int MACRO_LASTPORT = 65535;

  public AddrList addresses = new AddrList();

  // static info
  // ===========

  /**
   * Get static macro with given number
   * @param number Number of macro
   * @return macro
   */
  public static Macro staticMacro(int number) {
    Enumeration en = staticMacros.elements();
    Macro m;
    while (en.hasMoreElements()) {
      m = (Macro)en.nextElement();
      if (m.macroNumber == number)
        return m;
    }
    return null;
  }

  /**
   * Get static macro with given name
   * @param name name of macro
   * @return macro
   */
  public static Macro staticMacro(String name) {
    Enumeration en = staticMacros.elements();
    Macro m;
    while (en.hasMoreElements()) {
      m = (Macro)en.nextElement();
      if (m.macroName.equals(name))
        return m;
    }
    return null;
  }

  /**
   * Get macro with given name. First the static macros are searched and
   * then the dynamic macros are searched.
   * @param name Macro name
   * @param dynMacros Vector containing the dynamic macros
   */
  public static Macro getMacro(String name, Vector dynMacros) {
    Macro m = staticMacro(name);
    if (m == null) {
      Enumeration en = dynMacros.elements();
      while (en.hasMoreElements()) {
        m = (Macro)en.nextElement();
        if (m.macroName.equals(name))
          return m;
      }
    }
    else
      return m;
    return null;
  }

  /**
   * Get macro with given number. First the static macros are searched and
   * then the dynamic macros are searched.
   * @param number Macro number
   * @param dynMacros Vector containing the dynamic macros
   */
  public static Macro getMacro(int number, Vector dynMacros) {
    Macro m = staticMacro(number);
    if (m == null) {
      Enumeration en = dynMacros.elements();
      while (en.hasMoreElements()) {
        m = (Macro)en.nextElement();
        if (m.macroNumber == number)
          return m;
      }
    }
    else
      return m;
    return null;
  }

  /**
   * Construct list box containing all macros.
   * @param l List box
   * @param dynMacros Vector containing the dynamic macros
   * @param macronum Number of macro that is selected initially
   */
  public static void constructList(List l, Vector dynMacros, int macronum) {
    Enumeration en = staticMacros.elements();
    Macro m;
    while (en.hasMoreElements()) {
      m = (Macro)en.nextElement();
      l.add(m.macroName);
      if (m.macroNumber == macronum) {
        l.select(l.getItemCount()-1);
        l.makeVisible(l.getItemCount()-1);
      }
    }
    en = dynMacros.elements();
    while (en.hasMoreElements()) {
      m = (Macro)en.nextElement();
      l.add(m.macroName);
      if (m.macroNumber == macronum) {
        l.select(l.getItemCount()-1);
        l.makeVisible(l.getItemCount()-1);
      }
    }
  }

  /**
   * Construct list box containing all macros.
   * @param l List box
   * @param dynMacros Vector containing the dynamic macros
   */
  public static void constructList(List l, Vector dynMacros) {
    constructList(l, dynMacros, 0);
  }

  public static final int STATIC_INSIDE_OUTSIDE     = -100;
  public static final int STATIC_SERVER             = -101;

  public static final int STATIC_FIREWALLS          = -102;
  public static final int STATIC_FIREWALLS_UNPRIV   = -103;
  public static final int STATIC_FIREWALLS_CONFPORT = -104;
  public static final int STATIC_FIREWALLS_IDENTPORT= -105;
  public static final int STATIC_FIREWALLS_TRACEROUTEPORTS = -106;
  public static final int STATIC_CONFIG_CLIENTS     = -107;
  public static final int STATIC_INSIDE             = -108;
  public static final int STATIC_OUTSIDE            = -109;
  public static final int STATIC_LOCALHOST          = -110;
  public static final int STATIC_SMTPPORT           = -111;
  public static final int STATIC_IDENTPORT          = -112;
  public static final int STATIC_RUSERPORT          = -113;
  public static final int STATIC_DNSPORT            = -114;
  public static final int STATIC_FTPPORT            = -115;
  public static final int STATIC_TELNETPORT         = -116;
  public static final int STATIC_SSHPORT            = -117;
  public static final int STATIC_HTTPPORTS          = -118;
  public static final int STATIC_HTTPSPORTS         = -119;
  public static final int STATIC_TRACEROUTEPORTS    = -120;
  public static final int STATIC_INSIDE_IDENTPORT   = -121;
  public static final int STATIC_INSIDE_TRACEROUTEPORTS = -122;
  public static final int STATIC_OWNADDRESSES       = -123;
  public static final int STATIC_OWNADDRESSES_FWPORT = -124;
  public static final int STATIC_FINGERPORT         = -125;
  public static Vector staticMacros = new Vector(30,10);
  public static final int firstDynamic              = -200;

  static {
    Macro m;

    m = new Macro(STATIC_FIREWALLS);
    m.macroName = "FIREWALLS";
    m.macroType = MACRO_ADDRESSLIST;
    m.macroSpecial = MACRO_FIREWALLS;
    // addresses will be filled in automatically 
    staticMacros.addElement(m);

    m = new Macro(STATIC_FIREWALLS_UNPRIV);
    m.macroName = "FW-UNPRIV";
    m.macroType = MACRO_ADDRESSLIST;
    m.macroSpecial = MACRO_FIREWALLS;
    m.port = 1024;
    // addresses will be filled in automatically
    staticMacros.addElement(m);

    m = new Macro(STATIC_FIREWALLS_CONFPORT);
    m.macroName = "FW-CONFPORT";
    m.macroType = MACRO_ADDRESSLIST;
    m.macroSpecial = MACRO_FIREWALLS;
    m.port = 7227;
    m.prend = 7227;
    // addresses will be filled in automatically
    staticMacros.addElement(m);

    m = new Macro(STATIC_FIREWALLS_IDENTPORT);
    m.macroName = "FW-IDENTPORT";
    m.macroType = MACRO_ADDRESSLIST;
    m.macroSpecial = MACRO_FIREWALLS;
    m.port = 113;
    m.prend = 113;
    // addresses will be filled in automatically 
    staticMacros.addElement(m);

    m = new Macro(STATIC_FIREWALLS_TRACEROUTEPORTS);
    m.macroName = "FW-IDENTPORT";
    m.macroType = MACRO_ADDRESSLIST;
    m.macroSpecial = MACRO_FIREWALLS;
    m.port = 33434;
    m.prend = 33523;
    // addresses will be filled in automatically 
    staticMacros.addElement(m);

    m = new Macro(STATIC_CONFIG_CLIENTS);
    m.macroName = "CONFCLIENTS";
    m.macroType = MACRO_ADDRESSLIST;
    m.macroSpecial = MACRO_CONFCLIENTS;
    // addresses will be filled in automatically 
    staticMacros.addElement(m);

    m = new Macro(STATIC_INSIDE);
    m.macroName = "INSIDE";
    m.macroType = MACRO_INSIDE;
    staticMacros.addElement(m);

    m = new Macro(STATIC_OUTSIDE);
    m.macroName = "OUTSIDE";
    m.macroType = MACRO_OUTSIDE;
    staticMacros.addElement(m);

    m = new Macro(STATIC_LOCALHOST);
    m.macroName = "LOCALHOST";
    m.macroType = MACRO_ADDRESSLIST;
    m.addresses.addAddress("127.0.0.1");
    staticMacros.addElement(m);
    
    m = new Macro(STATIC_SMTPPORT);
    m.macroName = "SMTP-PORT";
    m.macroType = MACRO_PORTS_ONLY;
    m.port = 25;
    m.prend = 25;
    staticMacros.addElement(m);

    m = new Macro(STATIC_IDENTPORT);
    m.macroName = "IDENT-PORT";
    m.macroType = MACRO_PORTS_ONLY;
    m.port = 113;
    m.prend = 113;
    staticMacros.addElement(m);

    m = new Macro(STATIC_RUSERPORT);
    m.macroName = "RUSER-PORT";
    m.macroType = MACRO_PORTS_ONLY;
    m.port = 111;
    m.prend = 111;
    staticMacros.addElement(m);

    m = new Macro(STATIC_DNSPORT);
    m.macroName = "DNS-PORT";
    m.macroType = MACRO_PORTS_ONLY;
    m.port = 53;
    m.prend = 53;
    staticMacros.addElement(m);

    m = new Macro(STATIC_FTPPORT);
    m.macroName = "FTP-PORT";
    m.macroType = MACRO_PORTS_ONLY;
    m.port = 21;
    m.prend = 21;
    staticMacros.addElement(m);

    m = new Macro(STATIC_TELNETPORT);
    m.macroName = "TELNET-PORT";
    m.macroType = MACRO_PORTS_ONLY;
    m.port = 23;
    m.prend = 23;
    staticMacros.addElement(m);

    m = new Macro(STATIC_SSHPORT);
    m.macroName = "SSH-PORT";
    m.macroType = MACRO_PORTS_ONLY;
    m.port = 22;
    m.prend = 22;
    staticMacros.addElement(m);

    m = new Macro(STATIC_HTTPPORTS);
    m.macroName = "HTTP-PORTS";
    m.macroType = MACRO_PORTS_ONLY;
    m.port = 80;
    m.prend = 81;
    staticMacros.addElement(m);

    m = new Macro(STATIC_HTTPSPORTS);
    m.macroName = "HTTPS-PORTS";
    m.macroType = MACRO_PORTS_ONLY;
    m.port = 443;
    m.prend = 444;
    staticMacros.addElement(m);

    m = new Macro(STATIC_TRACEROUTEPORTS);
    m.macroName = "TRCRTE-PORTS";
    m.macroType = MACRO_PORTS_ONLY;
    m.port = 33434;
    m.prend = 33523;
    staticMacros.addElement(m);

    m = new Macro(STATIC_INSIDE_IDENTPORT);
    m.macroName = "INS_IDENT";
    m.macroType = MACRO_INSIDE;
    m.port = 113;
    m.prend = 113;
    staticMacros.addElement(m);

    m = new Macro(STATIC_INSIDE_TRACEROUTEPORTS);
    m.macroName = "INS_TRTE";
    m.macroType = MACRO_INSIDE;
    m.port = 33434;
    m.prend = 33523;
    staticMacros.addElement(m);

    m = new Macro(STATIC_OWNADDRESSES);
    m.macroName = "OWNADDR";
    m.macroType = MACRO_OWNADDR;
    staticMacros.addElement(m);

    m = new Macro(STATIC_OWNADDRESSES_FWPORT);
    m.macroName = "OWNFIREWALL";
    m.macroType = MACRO_OWNADDR;
    m.port = Communicator.serverPort;
    m.prend = Communicator.serverPort;
    staticMacros.addElement(m);

    m = new Macro(STATIC_FINGERPORT);
    m.macroName = "FINGER-PORT";
    m.macroType = MACRO_PORTS_ONLY;
    m.port = 79;
    m.prend = 79;
    staticMacros.addElement(m);

  }

} // Macro

// Dialog box for editing macros
class MacroDialog
	extends Dialog
	implements ActionListener {

  /**
   * @param co calling object, refresh() will be called after "OK"
   * @param p parent window
   * @param n macro to be edited
   * @param md manage domain holding dynamic macros
   * @param nw if true, macro will be removed from vector
   * in md when "Cancel" has been selected
   */   
  public MacroDialog(Refreshable co, Frame p, Macro m, ManageDomain md, boolean nw) {
    super(p, "Address Macros", true);

    callingObject = co;
    parent = p;
    macro = m;
    mgDomain = md;
    newMacro = nw;

    GridBagLayout gbl = new GridBagLayout();
    GridBagConstraints gbc = new GridBagConstraints();
    gbc.fill = GridBagConstraints.BOTH;
    gbc.anchor = GridBagConstraints.NORTHWEST;
    gbc.insets = new Insets(3,3,3,3);
    setLayout(gbl);

    mName = new TextField(m.macroName, 10);
    Utils.add_component(this, new Label("Name:"), gbl, gbc, 0, 0, 1, 1, 0, 0);
    Utils.add_component(this, mName, gbl, gbc, 1, 0, 1, 1, 0, 0);

    CheckboxGroup cg = new CheckboxGroup();
    insideBox = new Checkbox("Inside", cg, m.macroType == Macro.MACRO_INSIDE);
    Utils.add_component(this, insideBox, gbl, gbc, 0, 1, 2, 1, 0, 0);
    outsideBox = new Checkbox("Outside", cg, m.macroType == Macro.MACRO_OUTSIDE);
    Utils.add_component(this, outsideBox, gbl, gbc, 0, 2, 2, 1, 0, 0);
    addressBox = new Checkbox("Addresses", cg, m.macroType == Macro.MACRO_ADDRESSLIST);
    Utils.add_component(this, addressBox, gbl, gbc, 0, 3, 2, 1, 0, 0);
    portsonlyBox = new Checkbox("Ports only", cg, m.macroType == Macro.MACRO_PORTS_ONLY);
    Utils.add_component(this, portsonlyBox, gbl, gbc, 0, 4, 2, 1, 0, 0);

    Panel ap = m.addresses.editPanel();
    Utils.add_component(this, ap, gbl, gbc, 0, 5, 4, 8, 100, 100);

    Utils.add_component(this, new Label("Port"), gbl, gbc, 0, 13, 1, 1, 0, 0);
    portText = new TextField(6); // allow service names !!!
    Utils.add_component(this, portText, gbl, gbc, 1, 13, 1, 1, 50, 0);
    prendText = new TextField(6);
    if (m.port != Macro.MACRO_FIRSTPORT || m.prend != Macro.MACRO_LASTPORT) {
      portText.setText(Integer.toString(m.port));
      if (m.prend != m.port)
        prendText.setText(Integer.toString(m.prend));
    }
    Utils.add_component(this, new Label(".."), gbl, gbc, 2, 13, 1, 1, 0, 0);
    Utils.add_component(this, prendText, gbl, gbc, 3, 13, 1, 1, 50, 0);

    gbc.fill = GridBagConstraints.NONE;
    gbc.anchor = GridBagConstraints.WEST;
    Panel pa = new Panel();
    OKbutton = new Button("OK");
	 OKbutton.addActionListener(this);
    pa.add(OKbutton);
    // Cancelbutton = new Button("Cancel");
    // !!! pa.add(Cancelbutton); 
    Utils.add_component(this, pa, gbl, gbc, 0, 14, 4, 1, 0, 0);

    pack();
	 addWindowListener(new MDAdapter());
  }

class MDAdapter extends WindowAdapter {
  	public void windowClosing(WindowEvent we) {
    if (isEnabled())
        dispose();
    }
} // class MDAdapter


  public void actionPerformed(ActionEvent ae) {
  Object source = ae.getSource();

	 // Process OKbutton
    if (source == OKbutton) {

      // don't save changes in static macros
      if (macro.macroNumber <= Rule.SF_FIRSTMACRO && 
          macro.macroNumber > Macro.firstDynamic) {
        dispose();
        return;
      }

      String str = mName.getText().trim().toUpperCase();
      Enumeration en = mgDomain.Macros.elements();
      Macro m;
      while (en.hasMoreElements()) {
        m = (Macro)en.nextElement();
        if (m != macro && m.macroName.equals(str)) {
          UserDialog.ErrorBox("Macro "+str+" already exists!");
          return;
        }
      }
      macro.macroName = str;

      if (insideBox.getState())
        macro.macroType = Macro.MACRO_INSIDE;
      if (outsideBox.getState())
        macro.macroType = Macro.MACRO_OUTSIDE;
      if (portsonlyBox.getState())
        macro.macroType = Macro.MACRO_PORTS_ONLY;
      if (!addressBox.getState()) { 
        macro.addresses = new AddrList(); 
      }

      str = portText.getText().trim();
      if (str.length() > 0) { // !!! allow service names
        try {
          macro.port = Integer.parseInt(str);
          if (macro.port < Macro.MACRO_FIRSTPORT || macro.port > Macro.MACRO_LASTPORT)
            macro.port = Macro.MACRO_FIRSTPORT;
        }
        catch (NumberFormatException e) {
          macro.port = Macro.MACRO_FIRSTPORT;
        }
      }
      else
        macro.port = Macro.MACRO_FIRSTPORT;

      str = prendText.getText().trim();
      if (str.length() > 0) {
        try {
          macro.prend = Integer.parseInt(str);
          if (macro.prend < Macro.MACRO_FIRSTPORT || 
              macro.prend > Macro.MACRO_LASTPORT ||
              macro.prend < macro.port)
            macro.prend = macro.port == Macro.MACRO_FIRSTPORT ? Macro.MACRO_LASTPORT : macro.port;
        }
        catch (NumberFormatException e) {
          macro.prend = macro.port == Macro.MACRO_FIRSTPORT ? Macro.MACRO_LASTPORT : macro.port;
        }
      }
      else
        macro.prend = macro.port == Macro.MACRO_FIRSTPORT ? Macro.MACRO_LASTPORT : macro.port;

      if (callingObject != null)
        callingObject.refresh();
      dispose();
      return;
    }


	 // Process Cancelbutton
    else if (source == Cancelbutton) {
      if (newMacro)
        mgDomain.Macros.removeElement(macro);
      dispose();
      return;
    }
  } // actionPerformed


  private Refreshable callingObject;
  private Frame parent;
  private Macro macro;
  private ManageDomain mgDomain;
  private boolean newMacro;

  private Button OKbutton;
  private Button Cancelbutton;
  private TextField mName;
  private Checkbox insideBox;
  private Checkbox outsideBox;
  private Checkbox addressBox;
  private Checkbox portsonlyBox;
  private TextField portText;
  private TextField prendText;
} // MacroDialog
